# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import contextlib
import os
import time
from functools import partial
from typing import Any, Callable, cast, Dict, List, Optional, Tuple, TypeVar

import torch

from .common import BaseOperator, get_operator, get_xformers_operator, register_operator


@register_operator
class SparsifyBothWays(BaseOperator):
    OPERATOR = get_xformers_operator("sparse24_sparsify_both_ways")
    OPERATOR_CATEGORY = "sp24"
    NAME = "sparse24_sparsify_both_ways"


@register_operator
class SparsifyApply(BaseOperator):
    OPERATOR = get_xformers_operator("sparse24_apply")
    OPERATOR_CATEGORY = "sp24"
    NAME = "sparse24_apply"


@register_operator
class SparsifyApplyDenseOutput(BaseOperator):
    OPERATOR = get_xformers_operator("sparse24_apply_dense_output")
    OPERATOR_CATEGORY = "sp24"
    NAME = "sparse24_apply_dense_output"


@register_operator
class Sp24Gemm(BaseOperator):
    OPERATOR = get_xformers_operator("_sparse24_gemm")
    OPERATOR_CATEGORY = "sp24"
    NAME = "_sparse24_gemm"


def _get_cusparselt_torch_version() -> Tuple[int, int, int]:
    """
    Returns the version of the cusparselt.so library used by pytorch
    """
    if not torch.backends.cusparselt.is_available():
        return (0, 0, 0)
    version: Optional[int] = torch.backends.cusparselt.version()
    if version is None:
        return (0, 0, 0)
    return ((version // 10000) % 100, (version // 100) % 100, version % 100)


_cusplt_version = _get_cusparselt_torch_version()
_cusplt_version_str = ".".join(str(v) for v in _cusplt_version)


@register_operator
class Sp24GemmCuspltSearch(BaseOperator):
    OPERATOR = get_operator("aten", "_cslt_sparse_mm_search")
    OPERATOR_CATEGORY = "sp24"
    NAME = f"_cslt_sparse_mm_search@{_cusplt_version_str}"


@register_operator
class Sp24GemmCusplt(BaseOperator):
    OPERATOR = get_operator("aten", "_cslt_sparse_mm")
    OPERATOR_CATEGORY = "sp24"
    NAME = f"_cslt_sparse_mm@{_cusplt_version_str}"


def _has_cusparseLt() -> bool:
    available = _cusplt_version >= (0, 5, 0)
    if not available:
        return False

    # Sm90 added in 6.0
    compute_capability = (0, 0)
    if torch.cuda.is_available():
        compute_capability = torch.cuda.get_device_capability("cuda")
    if _cusplt_version < (6, 0, 0):
        if compute_capability >= (9, 0):
            return False
    return available


def sparse24_pointwise_op(
    func, types, args=(), kwargs=None, allow_sparsify_args_list=()
):
    self = None
    for tensor in args:
        if isinstance(tensor, Sparse24Tensor):
            self = tensor
    assert self is not None
    args_updated = []
    for i, tensor in enumerate(args):
        if isinstance(tensor, torch.Tensor):
            if not isinstance(tensor, Sparse24Tensor):
                if i in allow_sparsify_args_list:
                    tensor = sparsify24_like(tensor, self)
                else:
                    raise ValueError(
                        f"Operation {func.__module__}.{func.__name__} on Sparse24Tensor requires all operands to "
                        f"be Sparse24Tensors, but operand {i} is a {type(tensor)}"
                    )
            if (
                tensor.threads_masks is None
                or self.threads_masks is None
                or tensor.threads_masks.data_ptr() != self.threads_masks.data_ptr()
                or tensor.threads_masks.stride() != self.threads_masks.stride()
            ):
                raise ValueError(
                    f"Operation {func.__module__}.{func.__name__} on Sparse24Tensor requires all operands to be "
                    "Sparse24Tensors with the same sparsity pattern"
                )
        args_updated.append(tensor)
    assert isinstance(
        self, Sparse24TensorCutlass
    ), "Only implemented for CUTLASS tensors"
    return Sparse24TensorCutlass(
        self.shape,
        func(
            *[(x.packed if isinstance(x, Sparse24Tensor) else x) for x in args_updated]
        ),
        self.meta,
        func(
            *[
                (x.packed_t if isinstance(x, Sparse24Tensor) else x)
                for x in args_updated
            ]
        ),
        self.meta_t,
        self.threads_masks,
    )


def sparse24_mm(func, types, args=(), kwargs=None) -> torch.Tensor:
    assert len(args) == 2
    A, B = args
    if A.ndim != 2 or B.ndim != 2:
        raise NotImplementedError(
            "`Sparse24Tensor` matmul: Broadcasting is not implemented"
        )
    if isinstance(A, Sparse24Tensor):
        return A._mm(B)
    else:
        B_t = B.t()
        assert isinstance(B_t, Sparse24Tensor)
        return B_t._mm(A.t(), prefer_col_major_output=True).t()


def sparse24_addmm(func, types, args=(), kwargs=None) -> torch.Tensor:
    assert len(args) == 3
    bias, A, B = args
    if A.ndim != 2 or B.ndim != 2:
        raise NotImplementedError(
            "`Sparse24Tensor` matmul: Broadcasting is not implemented"
        )
    if bias.ndim != 1:
        raise NotImplementedError(
            f"`Sparse24Tensor` matmul: only bias dim=1 supported. Shape={bias.shape}"
        )
    if isinstance(A, Sparse24Tensor):
        raise NotImplementedError(
            "`Sparse24Tensor` matmul: only operand B of `addmm` can be sparse"
        )
    B_t = B.t()
    assert isinstance(B_t, Sparse24Tensor)
    return B_t._mm(A.t(), bias=bias, prefer_col_major_output=True).t()


def sparse24_linear(func, types, args=(), kwargs=None) -> torch.Tensor:
    assert len(args) in [2, 3]
    A, B = args[:2]
    bias = args[2] if len(args) == 3 else None
    if bias is None:
        return A @ B.t()
    return sparse24_addmm(
        func=None,
        types=None,
        args=[bias, A, B.t()],
    )


def sparse24_t(func, types, args=(), kwargs=None) -> torch.Tensor:
    assert len(args) == 1
    self = args[0]
    assert isinstance(self, Sparse24Tensor)
    assert len(self.shape) == 2
    return self.__class__(
        (self.shape[-1], self.shape[0]),
        packed=self.packed_t,
        meta=self.meta_t,
        packed_t=self.packed,
        meta_t=self.meta,
        threads_masks=self.threads_masks.transpose(0, 1),
    )


def sparse24_view(func, types, args=(), kwargs=None) -> torch.Tensor:
    assert len(args) == 2
    self, shape = args
    if tuple(shape) != self.shape:
        raise NotImplementedError(
            f"`view` is not implemented for Sparse24Tensor, except for the dummy case (shape={shape})"
        )
    return self


def sparse24_detach(func, types, args, kwargs) -> torch.Tensor:
    assert len(args) == 1
    self = args[0]
    return self.__class__(
        shape=self.shape,
        packed=self.packed,
        meta=self.meta,
        packed_t=self.packed_t,
        meta_t=self.meta_t,
        threads_masks=self.threads_masks,
        requires_grad=False,
    )


@contextlib.contextmanager
def no_dispatch():
    guard = torch._C._DisableTorchDispatch()
    try:
        yield
    finally:
        del guard


def fallback_dispatcher(func, types, args, kwargs):
    with no_dispatch():
        return func(*args)


SPARSE24_DISPATCH_CUTLASS = {
    torch.ops.aten.is_same_size: fallback_dispatcher,
    torch.ops.aten.detach_: fallback_dispatcher,
    torch.ops.aten.detach: sparse24_detach,
    torch.ops.aten.relu: sparse24_pointwise_op,
    torch.ops.aten.gelu: sparse24_pointwise_op,
    torch.ops.aten.silu: sparse24_pointwise_op,
    torch.ops.aten.mul: partial(
        # `mul` BW in swiglu
        sparse24_pointwise_op,
        allow_sparsify_args_list=(
            0,
            1,
        ),
    ),
    torch.ops.aten.add: sparse24_pointwise_op,
    # Note: for these ops, we allow the gradient to come in as a `torch.Tensor`
    # and we will run the sparsification right before calling the BW aten func
    torch.ops.aten.gelu_backward: partial(
        sparse24_pointwise_op, allow_sparsify_args_list=(0,)
    ),
    torch.ops.aten.silu_backward: partial(
        sparse24_pointwise_op, allow_sparsify_args_list=(0, 1)
    ),
    torch.ops.aten.threshold_backward: partial(  # relu BW
        sparse24_pointwise_op,
        allow_sparsify_args_list=(0,),
    ),
    torch.ops.aten.mm: sparse24_mm,
    torch.ops.aten.matmul: sparse24_mm,
    torch.ops.aten.t: sparse24_t,
    torch.ops.aten.view: sparse24_view,
    torch.ops.aten.linear: sparse24_linear,
}

SPARSE24_DISPATCH_CUSPARSELT = {
    torch.ops.aten.is_same_size: fallback_dispatcher,
    torch.ops.aten.detach_: fallback_dispatcher,
    torch.ops.aten.detach: sparse24_detach,
    torch.ops.aten.t: sparse24_t,
    torch.ops.aten.view: sparse24_view,
    torch.ops.aten.mm: sparse24_mm,
    torch.ops.aten.matmul: sparse24_mm,
    torch.ops.aten.addmm: sparse24_addmm,
    torch.ops.aten.linear: sparse24_linear,
}


class Sparse24Tensor(torch.Tensor):
    packed: torch.Tensor
    meta: torch.Tensor
    packed_t: torch.Tensor
    meta_t: torch.Tensor
    threads_masks: torch.Tensor
    __slots__ = ["packed", "meta", "packed_t", "meta_t", "threads_masks"]

    # We need to update the new method here to tell PyTorch what should be
    # the Tensor corresponding to the wrapper object
    @staticmethod
    def __new__(
        cls,
        shape,
        packed: torch.Tensor,
        meta: torch.Tensor,
        packed_t: torch.Tensor,
        meta_t: torch.Tensor,
        threads_masks: torch.Tensor,
        *,
        requires_grad=False,
    ):
        assert isinstance(packed, torch.Tensor)
        tensor = torch.Tensor._make_wrapper_subclass(  # type: ignore[attr-defined]
            cls,
            shape,
            device=packed.device,
            dtype=packed.dtype,
            requires_grad=requires_grad,
        )
        tensor.packed = packed
        tensor.meta = meta
        tensor.packed_t = packed_t
        tensor.meta_t = meta_t
        tensor.threads_masks = threads_masks
        return tensor

    def __repr__(self):
        return f"{self.__class__.__name__}(shape={self.shape})"

    def _sp24_to_dense(self) -> torch.Tensor:
        # Multiply by identity
        # WARN: This is not efficient at all
        e = torch.eye(
            self.shape[1], self.shape[1], device=self.device, dtype=self.dtype
        )
        return self @ e

    def _mm(
        self,
        B: torch.Tensor,
        *,
        prefer_col_major_output: bool = False,
        bias: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        raise NotImplementedError()

    __torch_function__ = torch._C._disabled_torch_function_impl

    def __tensor_flatten__(self):
        return self.__slots__, (self.shape, self.requires_grad)

    @classmethod
    def __tensor_unflatten__(
        cls, inner_tensors, flatten_spec, outer_size, outer_stride
    ):
        shape, requires_grad = flatten_spec
        return cls(
            shape,
            **inner_tensors,
            requires_grad=requires_grad,
        )


class Sparse24TensorCutlass(Sparse24Tensor):
    def _mm(
        self,
        B: torch.Tensor,
        *,
        bias: Optional[torch.Tensor] = None,
        prefer_col_major_output: bool = False,
    ) -> torch.Tensor:
        if isinstance(B, Sparse24Tensor):
            raise ValueError(
                "`Sparse24Tensor @ Sparse24Tensor` is not supported by the hardware"
            )
        if bias is not None:
            raise NotImplementedError(
                f"`Sparse24Tensor` with backend='{BACKEND_CUTLASS}' does not support matmul with bias. "
                f"Remove the bias, or use backend='{BACKEND_CUSPARSELT}'"
            )
        if self.ndim != 2 or B.ndim != 2:
            raise NotImplementedError(
                f"`{self.__class__.__name__}` matmul: Broadcasting is not implemented"
            )
        if self.shape[1] != B.shape[0]:
            raise NotImplementedError(
                f"`{self.__class__.__name__}` matmul: invalid shapes \
    ({self.shape[0]}, {self.shape[1]}) @ ({B.shape[0]}, {B.shape[1]})"
            )
        return Sp24Gemm.OPERATOR(self.packed, B, self.meta)[: self.shape[0]]

    @classmethod
    def __torch_dispatch__(cls, func, types, args=(), kwargs=None):
        if func._overloadpacket not in SPARSE24_DISPATCH_CUTLASS:
            raise NotImplementedError(
                f"{cls.__name__} only supports a specific set of operations, "
                f"can't perform requested op ({func.__name__})"
            )
        return SPARSE24_DISPATCH_CUTLASS[func._overloadpacket](
            func, types, args, kwargs
        )


_CUSPLT_ALG_CACHE: Dict[Tuple[int, int, int, str, torch.dtype, bool], int] = {}
# Disabled by default, as there is a correctness issue on cusparselt
# when using some algorithms:
# https://github.com/pytorch/pytorch/issues/155333
_CUSPLT_TUNE = os.environ.get("XFORMERS_CUSPARSELT_TUNE", "0") == "1"


def _cusplt_find_alg(
    shape: List[int],
    packed: torch.Tensor,
    B: torch.Tensor,
    bias: Optional[torch.Tensor],
    transpose_result: bool,
) -> int:
    """
    cuSPARSELt has multiple algorithms (that correspond to different kernels)
    to run a given GEMM, because the optimal kernel depends on the GEMM dimensions.
    This function attempts to find the most efficient one by benchmarking all
    of them.
    NOTE: cuSPARSELt also provides a function to search the best algorithm
    (exposed via `aten:_cslt_sparse_mm_search`) but it often fails to find the best
    algorithm, so we need this workaround.
    """
    if not _CUSPLT_TUNE:
        return 0
    M, K = shape
    N = B.shape[1]
    fmt = "r"
    fmt += "r" if B.stride(-1) <= 1 else "c"
    fmt += "c" if transpose_result else "r"
    h = (M, N, K, fmt, B.dtype, bias is not None)
    if h in _CUSPLT_ALG_CACHE:
        return _CUSPLT_ALG_CACHE[h]

    REPEAT = 10
    TIME_ALGO = []
    for algo in range(70):
        has_error = False
        for i in range(REPEAT):
            try:
                Sp24GemmCusplt.OPERATOR(
                    packed, B, bias=bias, transpose_result=transpose_result, alg_id=algo
                )
            except RuntimeError:
                has_error = True
                break
            if i == 1:  # 1 iteration of warmup
                torch.cuda.synchronize()
                t = time.monotonic()
        if has_error:
            break
        torch.cuda.synchronize()
        dt = time.monotonic() - t
        TIME_ALGO.append((dt, algo))
    TIME_ALGO.sort()
    _CUSPLT_ALG_CACHE[h] = TIME_ALGO[0][1]
    return _CUSPLT_ALG_CACHE[h]


@torch.library.custom_op("xformers::_cusplt_mm", mutates_args=(), device_types=["cuda"])
def _cusplt_mm(
    shape: List[int],
    packed: torch.Tensor,
    B: torch.Tensor,
    bias: Optional[torch.Tensor],
    transpose_result: bool,
) -> torch.Tensor:
    """
    This operator wraps find_algo + gemm. This is because we don't want find_algo
    to be visible by torch compile, otherwise it will remove it from the graph.
    """
    alg_id = _cusplt_find_alg(
        shape, packed, B, bias=bias, transpose_result=transpose_result
    )
    return Sp24GemmCusplt.OPERATOR(
        packed, B, bias=bias, transpose_result=transpose_result, alg_id=alg_id
    )


@torch.library.register_fake("xformers::_cusplt_mm")
def _cusplt_mm_meta(
    shape: List[int],
    packed: torch.Tensor,
    B: torch.Tensor,
    bias: Optional[torch.Tensor],
    transpose_result: bool,
) -> torch.Tensor:
    M, K = shape
    N = B.shape[1]
    if transpose_result:
        return torch.empty([N, M], dtype=B.dtype, device=B.device)
    return torch.empty([M, N], dtype=B.dtype, device=B.device)


class Sparse24TensorCuSparseLt(Sparse24Tensor):
    def _mm(
        self,
        B: torch.Tensor,
        *,
        prefer_col_major_output: bool = False,
        bias: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if isinstance(B, Sparse24Tensor):
            raise ValueError(
                "`Sparse24Tensor @ Sparse24Tensor` is not supported by the hardware"
            )
        if self.ndim != 2 or B.ndim != 2:
            raise NotImplementedError(
                f"`{self.__class__.__name__}` matmul: Broadcasting is not implemented"
            )
        if self.shape[1] != B.shape[0]:
            raise NotImplementedError(
                f"`{self.__class__.__name__}` matmul: invalid shapes \
    ({self.shape[0]}, {self.shape[1]}) @ ({B.shape[0]}, {B.shape[1]})"
            )
        if B.shape[1] % 8 != 0:
            raise NotImplementedError(
                f"`{self.__class__.__name__}` matmul: trying to do `A={tuple(self.shape)} @ B={tuple(B.shape)}`. "
                "The dense matrix B should have the second dimension aligned to 8."
            )
        if B.dtype != self.dtype:
            raise NotImplementedError(
                f"`{self.__class__.__name__}` matmul: trying to do `A={tuple(self.shape)} @ B={tuple(B.shape)}`, "
                f"with A.dtype={self.dtype} and B.dtype={B.dtype}. "
                "This operation is only supported when A and B have the same data type."
            )
        if bias is not None and bias.dtype != self.dtype:
            raise NotImplementedError(
                f"`{self.__class__.__name__}` matmul: trying to do `A={tuple(self.shape)} @ B={tuple(B.shape)} + C`, "
                f"with A.dtype=B.dtype={self.dtype} and C.dtype={B.dtype}. "
                "This operation is only supported when A, B and C have the same data type."
            )
        assert _has_cusparseLt()
        out = torch.ops.xformers._cusplt_mm(
            self.shape,
            self.packed,
            B,
            bias=bias,
            transpose_result=prefer_col_major_output,
        )
        if prefer_col_major_output:
            out = out.t()
        return out[: self.shape[0]]

    @classmethod
    def __torch_dispatch__(cls, func, types, args=(), kwargs=None):
        if func._overloadpacket not in SPARSE24_DISPATCH_CUSPARSELT:
            raise NotImplementedError(
                f"{cls.__name__} only supports a specific set of operations, "
                f"can't perform requested op ({func.__name__})"
            )
        return SPARSE24_DISPATCH_CUSPARSELT[func._overloadpacket](
            func, types, args, kwargs
        )


torch._dynamo.allow_in_graph(Sparse24TensorCuSparseLt)
torch._dynamo.allow_in_graph(Sparse24TensorCutlass)

GRADIENT_SP24 = "24sparse"
GRADIENT_DENSE = "24dense"
GRADIENT_STE = "ste"  # Straight-Through Estimator

BACKEND_CUTLASS = "cutlass"
BACKEND_CUSPARSELT = "cusparselt"
BACKEND_DENSE = "dense"


def _sparsify24_forward(x: torch.Tensor, *, algo: str, backend: str) -> Sparse24Tensor:
    assert backend in [
        BACKEND_CUTLASS,
        BACKEND_CUSPARSELT,
    ], f"Invalid backend: {backend}"
    if isinstance(x, Sparse24Tensor):
        if x.threads_masks is None:
            raise ValueError("Input to `sparsify24` is already sparse")
        return x

    (packed, meta, packed_t, meta_t, threads_masks) = SparsifyBothWays.OPERATOR(
        x, algorithm=algo, backend=backend
    )
    cls = (
        Sparse24TensorCutlass
        if backend == BACKEND_CUTLASS
        else Sparse24TensorCuSparseLt
    )
    return cls(
        x.shape,
        packed=packed,
        meta=meta,
        packed_t=packed_t,
        meta_t=meta_t,
        threads_masks=threads_masks,
        requires_grad=False,
    )


class _Sparsify24Func(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x: torch.Tensor, algo: str, gradient: str, backend: str):  # type: ignore[override]
        if gradient not in [GRADIENT_SP24, GRADIENT_DENSE, GRADIENT_STE]:
            raise ValueError(
                f"Invalid gradient type: '{gradient}'. "
                f"Expected '{GRADIENT_SP24}' or '{GRADIENT_DENSE}' or '{GRADIENT_STE}"
            )
        out = _sparsify24_forward(x, algo=algo, backend=backend)
        ctx.threads_masks = out.threads_masks
        ctx.meta = out.meta
        ctx.meta_t = out.meta_t
        ctx.dtype = out.dtype
        ctx.gradient = gradient
        return out

    @staticmethod
    def backward(ctx, grad_out: torch.Tensor):  # type: ignore[override]
        if isinstance(grad_out, Sparse24Tensor) or ctx.gradient == GRADIENT_STE:
            return grad_out, None, None, None
        assert not isinstance(grad_out, Sparse24Tensor)
        assert grad_out.dtype == ctx.dtype
        if ctx.gradient == GRADIENT_SP24:
            packed, _, packed_t, _ = SparsifyApply.OPERATOR(grad_out, ctx.threads_masks)
            grad_in: torch.Tensor = Sparse24TensorCutlass(
                grad_out.shape,
                packed,
                ctx.meta,
                packed_t,
                ctx.meta_t,
                ctx.threads_masks,
                requires_grad=grad_out.requires_grad,
            )
        elif ctx.gradient == GRADIENT_DENSE:
            assert ctx.threads_masks.is_contiguous()
            grad_in = SparsifyApplyDenseOutput.OPERATOR(grad_out, ctx.threads_masks)
        else:
            assert False, f"Unsupported gradient type: {ctx.gradient}"
        return (
            grad_in,
            None,
            None,
            None,
        )


class _Sparsify24STEFunc(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        x: torch.Tensor,
        algo: str,
        backend: str,
        bw_mul0: float,
        bw_mul1: float,
    ):  # type: ignore[override]
        out = _sparsify24_forward(x, algo=algo, backend=backend)
        ctx.threads_masks = out.threads_masks
        ctx.bw_mul0 = bw_mul0
        ctx.bw_mul1 = bw_mul1
        return out

    @staticmethod
    def backward(ctx, grad_out: torch.Tensor):  # type: ignore[override]
        assert not isinstance(grad_out, Sparse24Tensor)
        if ctx.bw_mul0 == 1.0 and ctx.bw_mul1 == 1.0:
            grad_in = grad_out
        else:
            grad_in = SparsifyApplyDenseOutput.OPERATOR(
                grad_out, ctx.threads_masks, mul0=ctx.bw_mul0, mul1=ctx.bw_mul1
            )
        return (
            grad_in,
            None,  # algo
            None,  # backend
            None,  # bw_mul0
            None,  # bw_mul1
        )


class _Sparsify24LikeFunc(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x: torch.Tensor, pattern: Sparse24Tensor, gradient: str, backend: str):  # type: ignore[override]
        if not isinstance(pattern, Sparse24Tensor):
            raise NotImplementedError(
                "`sparsify24_like`: `pattern` must be a sparse tensor"
            )
        if not pattern.threads_masks.is_contiguous():
            raise NotImplementedError(
                "`sparsify24_like` is not implemented when `pattern` is transposed"
            )
        if gradient not in [GRADIENT_DENSE, GRADIENT_SP24, GRADIENT_STE]:
            raise ValueError(f'`sparsify24_like`: invalid gradient type "{gradient}"')
        ctx.threads_masks = pattern.threads_masks
        ctx.meta = pattern.meta
        ctx.meta_t = pattern.meta_t
        ctx.dtype = pattern.dtype
        ctx.gradient = gradient
        if backend == BACKEND_DENSE:
            assert ctx.threads_masks.is_contiguous()
            return SparsifyApplyDenseOutput.OPERATOR(x, ctx.threads_masks)
        packed, meta, packed_t, meta_t = SparsifyApply.OPERATOR(
            x, ctx.threads_masks, backend=backend
        )
        if backend == BACKEND_CUTLASS:
            return Sparse24TensorCutlass(
                x.shape,
                packed,
                ctx.meta,
                packed_t,
                ctx.meta_t,
                ctx.threads_masks,
                requires_grad=x.requires_grad,
            )
        assert backend == BACKEND_CUSPARSELT, f"Invalid backend: {backend}"
        meta.copy_(pattern.meta)
        meta_t.copy_(pattern.meta_t)
        return Sparse24TensorCuSparseLt(
            x.shape,
            packed,
            meta,
            packed_t,
            meta_t,
            ctx.threads_masks,
            requires_grad=x.requires_grad,
        )

    @staticmethod
    def backward(ctx, grad_out: torch.Tensor):  # type: ignore[override]
        if ctx.gradient == GRADIENT_STE or isinstance(grad_out, Sparse24Tensor):
            return grad_out, None, None, None
        assert not isinstance(grad_out, Sparse24Tensor)
        assert grad_out.dtype == ctx.dtype

        if ctx.gradient == GRADIENT_DENSE:
            assert ctx.threads_masks.is_contiguous()
            return (
                SparsifyApplyDenseOutput.OPERATOR(grad_out, ctx.threads_masks),
                None,
                None,
                None,
            )
        assert ctx.gradient == GRADIENT_SP24

        packed, _, packed_t, _ = SparsifyApply.OPERATOR(
            grad_out, ctx.threads_masks, backend=BACKEND_CUTLASS
        )
        return (
            Sparse24TensorCutlass(
                grad_out.shape,
                packed,
                ctx.meta,
                packed_t,
                ctx.meta_t,
                ctx.threads_masks,
                requires_grad=grad_out.requires_grad,
            ),
            None,
            None,
            None,
        )


# We want to use `torch._dynamo.allow_in_graph` as a decorator
# (see https://fburl.com/workplace/uimiz0mf) but it breaks mypy.
# This is a hack to work around this
F = TypeVar("F", bound=Callable[..., Any])


def allow_in_graph(func: F) -> F:
    return cast(F, torch._dynamo.allow_in_graph(func))


@allow_in_graph
def sparsify24(
    x: torch.Tensor,
    algo: str = "",
    gradient: str = GRADIENT_SP24,
    backend: str = BACKEND_CUTLASS,
) -> Sparse24Tensor:
    return _Sparsify24Func.apply(x, algo, gradient, backend)


@allow_in_graph
def sparsify24_ste(
    x: torch.Tensor,
    algo: str = "",
    backend: str = BACKEND_CUTLASS,
    bw_mul0: float = 1.0,
    bw_mul1: float = 1.0,
) -> Sparse24Tensor:
    """
    2:4 sparsification, with Straight Through Estimator for the
    backward pass (eg the gradient is *not* sparsified).
    Optionally, `bw_mul[0-1]` provide the option to rescale the gradient
    differently for pruned (`bw_mul0`) and kept values (`bw_mul1`).
    """
    return _Sparsify24STEFunc.apply(x, algo, backend, bw_mul0, bw_mul1)


@allow_in_graph
def sparsify24_like(
    x: torch.Tensor,
    pattern: torch.Tensor,
    gradient: str = GRADIENT_SP24,
    backend: str = "",
    out_dense: Optional[bool] = None,  # <-- TODO: Deprecate this in favor of "gradient"
) -> Sparse24Tensor:
    if out_dense is not None and out_dense:
        backend = BACKEND_DENSE
    if backend == "":
        backend = (
            BACKEND_CUSPARSELT
            if isinstance(pattern, Sparse24TensorCuSparseLt)
            else BACKEND_CUTLASS
        )
    if not isinstance(pattern, Sparse24Tensor):
        raise ValueError(
            f"`pattern` must be a `Sparse24Tensor` but got a {type(pattern)}"
        )
    # Handle transposed case
    if not pattern.threads_masks.is_contiguous():
        return _Sparsify24LikeFunc.apply(x.t(), pattern.t(), gradient, backend).t()
    return _Sparsify24LikeFunc.apply(x, pattern, gradient, backend)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import itertools
from typing import List, Tuple

import torch
import triton
import triton.language as tl

from xformers.ops._triton.matmul_perf_model import (
    early_config_prune,
    estimate_matmul_time,
)


def init_to_zero(*names):
    def result(nargs):
        for name in names:
            nargs[name].zero_()

    return result


def gen_config(
    block_m: int,
    block_n: int,
    block_k: int,
    stages: int,
    warps: int,
    split_k: int = 1,
    group_m: int = 8,
) -> triton.Config:
    """A more compact way to define a triton.Config, so it fits on one line"""

    return triton.Config(
        {
            "BLOCK_M": block_m,
            "BLOCK_N": block_n,
            "BLOCK_K": block_k,
            "SPLIT_K": split_k,
            "GROUP_M": group_m,
        },
        num_stages=stages,
        num_warps=warps,
        pre_hook=(
            init_to_zero(*[f"C{i+1}{j+1}" for i in range(3) for j in range(3)])
            if split_k > 1
            else init_to_zero()
        ),
    )


BASIC_MATMUL_CONFIGS = [
    gen_config(block_m=128, block_n=256, block_k=32, stages=3, warps=8),
    gen_config(block_m=256, block_n=128, block_k=32, stages=3, warps=8),
    gen_config(block_m=256, block_n=64, block_k=32, stages=4, warps=4),
    gen_config(block_m=64, block_n=256, block_k=32, stages=4, warps=4),
    gen_config(block_m=128, block_n=128, block_k=32, stages=4, warps=4),
    gen_config(block_m=128, block_n=64, block_k=32, stages=4, warps=4),
    gen_config(block_m=64, block_n=128, block_k=32, stages=4, warps=4),
    gen_config(block_m=128, block_n=32, block_k=32, stages=4, warps=4),
    gen_config(block_m=64, block_n=32, block_k=32, stages=5, warps=2),
]


INT8_MATMUL_CONFIGS = [
    gen_config(block_m=128, block_n=256, block_k=128, stages=3, warps=8),
    gen_config(block_m=256, block_n=128, block_k=128, stages=3, warps=8),
    gen_config(block_m=256, block_n=64, block_k=128, stages=4, warps=4),
    gen_config(block_m=64, block_n=256, block_k=128, stages=4, warps=4),
    gen_config(block_m=128, block_n=128, block_k=128, stages=4, warps=4),
    gen_config(block_m=128, block_n=64, block_k=64, stages=4, warps=4),
    gen_config(block_m=64, block_n=128, block_k=64, stages=4, warps=4),
    gen_config(block_m=128, block_n=32, block_k=64, stages=4, warps=4),
    gen_config(block_m=64, block_n=32, block_k=64, stages=5, warps=2),
]


IO_BOUND_MATMUL_CONFIGS_STAGES = [2, 3, 4, 5, 6]
IO_BOUND_MATMUL_CONFIGS_BLOCK_M = [16, 32]
IO_BOUND_MATMUL_CONFIGS_BLOCK_K = [32, 64]
IO_BOUND_MATMUL_CONFIGS_BLOCK_N = [32, 64, 128, 256]
IO_BOUND_MATMUL_CONFIGS_SPLIT_K = [1, 2, 4, 8, 16]


IO_BOUND_MATMUL_CONFIGS = [
    gen_config(
        block_m=block_m,
        block_n=block_n,
        block_k=block_k,
        stages=stages,
        warps=2 if block_n <= 64 else 4,
        split_k=split_k,
    )
    for stages, block_m, block_k, block_n, split_k in itertools.product(
        IO_BOUND_MATMUL_CONFIGS_STAGES,
        IO_BOUND_MATMUL_CONFIGS_BLOCK_M,
        IO_BOUND_MATMUL_CONFIGS_BLOCK_K,
        IO_BOUND_MATMUL_CONFIGS_BLOCK_N,
        IO_BOUND_MATMUL_CONFIGS_SPLIT_K,
    )
]


TRITON_CONFIGS = BASIC_MATMUL_CONFIGS + INT8_MATMUL_CONFIGS + IO_BOUND_MATMUL_CONFIGS


def our_estimate_matmul_time(
    A11, B11, C11, M1, M2, M3, N1, N2, N3, K1, K2, K3, **kwargs
):
    """Call into Triton's upstream cost model, with the right args

    The upstream function expects arguments to have certain names. Since we
    renamed a few of them in our implementation, we rename them back.

    At the time of writing (July 2023) the arguments that Triton expects are:
    M, N, K, A, B, C, BLOCK_M, BLOCK_N, BLOCK_K, SPLIT_K, num_warps, num_stages.

    """
    return estimate_matmul_time(
        M=M1 + M2 + M3, N=N1 + N2 + N3, K=K1 + K2 + K3, A=A11, B=B11, C=C11, **kwargs
    )


def our_early_config_prune(config, named_args, **kwargs):
    new_named_args = named_args.copy()
    new_named_args["M"] = named_args["M1"] + named_args["M2"] + named_args["M3"]
    new_named_args["N"] = named_args["N1"] + named_args["N2"] + named_args["N3"]
    new_named_args["K"] = named_args["K1"] + named_args["K2"] + named_args["K3"]
    new_named_args["A"] = named_args["A11"]
    new_named_args["B"] = named_args["B11"]
    new_named_args["C"] = named_args["C11"]
    return early_config_prune(config, new_named_args, **kwargs)


@triton.autotune(
    configs=TRITON_CONFIGS,
    key=["M1", "M2", "M3", "N1", "N2", "N3", "K1", "K2", "K3"],
    prune_configs_by={
        "early_config_prune": our_early_config_prune,
        "perf_model": our_estimate_matmul_time,
        "top_k": 10,
    },
)
@triton.heuristics(
    {
        "EVEN_K": lambda args: all(
            k % (args["BLOCK_K"] * args["SPLIT_K"]) == 0
            for k in [args["K1"], args["K2"], args["K3"]]
        ),
    }
)
@triton.jit()
def _xformers_tiled_matmul_kernel(
    A11,
    A12,
    A13,
    A21,
    A22,
    A23,
    A31,
    A32,
    A33,
    B11,
    B12,
    B13,
    B21,
    B22,
    B23,
    B31,
    B32,
    B33,
    C11,
    C12,
    C13,
    C21,
    C22,
    C23,
    C31,
    C32,
    C33,
    M1,
    M2,
    M3,
    N1,
    N2,
    N3,
    K1,
    K2,
    K3,
    stride_am1,
    stride_am2,
    stride_am3,
    stride_ak1,
    stride_ak2,
    stride_ak3,
    stride_bk1,
    stride_bk2,
    stride_bk3,
    stride_bn1,
    stride_bn2,
    stride_bn3,
    stride_cm1,
    stride_cm2,
    stride_cm3,
    stride_cn1,
    stride_cn2,
    stride_cn3,
    BLOCK_M: tl.constexpr,  # DO NOT CHANGE NAME: MUST MATCH PERF MODEL
    BLOCK_N: tl.constexpr,  # DO NOT CHANGE NAME: MUST MATCH PERF MODEL
    BLOCK_K: tl.constexpr,  # DO NOT CHANGE NAME: MUST MATCH PERF MODEL
    GROUP_M: tl.constexpr,
    SPLIT_K: tl.constexpr,  # DO NOT CHANGE NAME: MUST MATCH PERF MODEL
    EVEN_K: tl.constexpr,
    ACC_TYPE: tl.constexpr,
):
    # matrix multiplication
    pid = tl.program_id(0)
    pid_k = tl.program_id(1)
    grid_m1 = tl.cdiv(M1, BLOCK_M)
    grid_m2 = tl.cdiv(M2, BLOCK_M)
    grid_m3 = tl.cdiv(M3, BLOCK_M)
    grid_n1 = tl.cdiv(N1, BLOCK_N)
    grid_n2 = tl.cdiv(N2, BLOCK_N)
    grid_n3 = tl.cdiv(N3, BLOCK_N)
    grid_m = grid_m1 + grid_m2 + grid_m3
    grid_n = grid_n1 + grid_n2 + grid_n3

    # re-order program ID for better L2 performance
    width = GROUP_M * grid_n
    group_id = pid // width
    group_size = min(grid_m - group_id * GROUP_M, GROUP_M)
    pid_m = group_id * GROUP_M + (pid % group_size)
    pid_n = (pid % width) // (group_size)

    # We use tl.where to circumvent a regression in alignment auto-detection:
    # https://github.com/openai/triton/issues/1784

    A1 = tl.where(pid_m < grid_m1, A11, tl.where(pid_m < grid_m1 + grid_m2, A21, A31))
    A2 = tl.where(pid_m < grid_m1, A12, tl.where(pid_m < grid_m1 + grid_m2, A22, A32))
    A3 = tl.where(pid_m < grid_m1, A13, tl.where(pid_m < grid_m1 + grid_m2, A23, A33))
    B1 = tl.where(pid_n < grid_n1, B11, tl.where(pid_n < grid_n1 + grid_n2, B12, B13))
    B2 = tl.where(pid_n < grid_n1, B21, tl.where(pid_n < grid_n1 + grid_n2, B22, B23))
    B3 = tl.where(pid_n < grid_n1, B31, tl.where(pid_n < grid_n1 + grid_n2, B32, B33))
    C = tl.where(
        pid_m < grid_m1,
        tl.where(pid_n < grid_n1, C11, tl.where(pid_n < grid_n1 + grid_n2, C12, C13)),
        tl.where(
            pid_m < grid_m1 + grid_m2,
            tl.where(
                pid_n < grid_n1, C21, tl.where(pid_n < grid_n1 + grid_n2, C22, C23)
            ),
            tl.where(
                pid_n < grid_n1, C31, tl.where(pid_n < grid_n1 + grid_n2, C32, C33)
            ),
        ),
    )
    M = tl.where(pid_m < grid_m1, M1, tl.where(pid_m < grid_m1 + grid_m2, M2, M3))
    N = tl.where(pid_n < grid_n1, N1, tl.where(pid_n < grid_n1 + grid_n2, N2, N3))
    stride_ak = tl.where(
        pid_m < grid_m1,
        stride_ak1,
        tl.where(pid_m < grid_m1 + grid_m2, stride_ak2, stride_ak3),
    )
    stride_bk = tl.where(
        pid_n < grid_n1,
        stride_bk1,
        tl.where(pid_n < grid_n1 + grid_n2, stride_bk2, stride_bk3),
    )
    stride_cn = tl.where(
        pid_m < grid_m1,
        stride_cn1,
        tl.where(pid_m < grid_m1 + grid_m2, stride_cn2, stride_cn3),
    )
    stride_cm = tl.where(
        pid_n < grid_n1,
        stride_cm1,
        tl.where(pid_n < grid_n1 + grid_n2, stride_cm2, stride_cm3),
    )
    pid_m = tl.where(
        pid_m < grid_m1,
        pid_m,
        tl.where(pid_m < grid_m1 + grid_m2, pid_m - grid_m1, pid_m - grid_m1 - grid_m2),
    )
    pid_n = tl.where(
        pid_n < grid_n1,
        pid_n,
        tl.where(pid_n < grid_n1 + grid_n2, pid_n - grid_n1, pid_n - grid_n1 - grid_n2),
    )

    # do matrix multiplication
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    ram = tl.max_contiguous(tl.multiple_of(rm % M, BLOCK_M), BLOCK_M)
    rbn = tl.max_contiguous(tl.multiple_of(rn % N, BLOCK_N), BLOCK_N)
    # pointers
    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=ACC_TYPE)
    grid_k1 = tl.cdiv(K1, BLOCK_K)
    grid_k2 = tl.cdiv(K2, BLOCK_K)
    grid_k3 = tl.cdiv(K3, BLOCK_K)
    for tile in range(pid_k, grid_k1 + grid_k2 + grid_k3, SPLIT_K):
        A = tl.where(tile < grid_k1, A1, tl.where(tile < grid_k1 + grid_k2, A2, A3))
        B = tl.where(tile < grid_k1, B1, tl.where(tile < grid_k1 + grid_k2, B2, B3))
        K = tl.where(tile < grid_k1, K1, tl.where(tile < grid_k1 + grid_k2, K2, K3))
        stride_am = tl.where(
            tile < grid_k1,
            stride_am1,
            tl.where(tile < grid_k1 + grid_k2, stride_am2, stride_am3),
        )
        stride_bn = tl.where(
            tile < grid_k1,
            stride_bn1,
            tl.where(tile < grid_k1 + grid_k2, stride_bn2, stride_bn3),
        )
        my_tile = tl.where(
            tile < grid_k1,
            tile,
            tl.where(
                tile < grid_k1 + grid_k2, tile - grid_k1, tile - grid_k1 - grid_k2
            ),
        )
        rk = my_tile * BLOCK_K + tl.arange(0, BLOCK_K)
        Ain = A + (ram[:, None] * stride_am + rk[None, :] * stride_ak)
        Bin = B + (rk[:, None] * stride_bk + rbn[None, :] * stride_bn)
        if EVEN_K:
            a = tl.load(Ain)
            b = tl.load(Bin)
        else:
            a = tl.load(Ain, mask=rk[None, :] < K, other=0.0)
            b = tl.load(Bin, mask=rk[:, None] < K, other=0.0)
        acc += tl.dot(a, b, allow_tf32=False)
    acc = acc.to(C.dtype.element_ty)
    # rematerialize rm and rn to save registers
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    C = C + (rm[:, None] * stride_cm + rn[None, :] * stride_cn)
    mask = (rm < M)[:, None] & (rn < N)[None, :]
    # handles write-back with reduction-splitting
    if SPLIT_K == 1:
        tl.store(C, acc, mask=mask)
    else:
        tl.atomic_add(C, acc, mask=mask)


def _check_row_or_column(row_or_col_type, row_or_col_idx, tensor_name, dim_name, vals):
    assert len(vals) > 0
    for pos, val in enumerate(vals[1:]):
        assert val == vals[0], (
            f"the tensors on {row_or_col_type} {row_or_col_idx} of the {tensor_name} "
            f"must all have the same stride along the {dim_name} dimension, got "
            f"{vals[0]} at position 0 and {val} at position {pos + 1}"
        )
    return vals[0]


def _get_strides(
    ts: List[List[torch.Tensor]], tensor_name, dim_0_name, dim_1_name
) -> Tuple[List[int], List[int]]:
    strides_0 = [
        _check_row_or_column(
            "column", idx, tensor_name, dim_0_name, [y.stride(0) for y in x]
        )
        for idx, x in enumerate(zip(*ts))
    ]
    strides_1 = [
        _check_row_or_column(
            "row", idx, tensor_name, dim_1_name, [y.stride(1) for y in x]
        )
        for idx, x in enumerate(ts)
    ]
    assert all(s == 1 for s in strides_0) or all(s == 1 for s in strides_1)
    while len(strides_0) < 3:
        strides_0.append(1 if strides_0[0] == 1 else 0)
    while len(strides_1) < 3:
        strides_1.append(1 if strides_1[0] == 1 else 0)
    return strides_0, strides_1


def _launch_triton_matmul(
    a: List[List[torch.Tensor]],
    b: List[List[torch.Tensor]],
    c: List[List[torch.Tensor]],
    ms: List[int],
    ns: List[int],
    ks: List[int],
) -> None:
    strides_am, strides_ak = _get_strides(a, "first operand", "m", "k")
    strides_bk, strides_bn = _get_strides(b, "second operand", "k", "n")
    strides_cm, strides_cn = _get_strides(c, "output", "m", "n")

    # accumulator types
    ACC_TYPE = (
        tl.float32
        if c[0][0].dtype in [torch.float16, torch.bfloat16, torch.float32]
        else tl.int32
    )

    # launch kernel
    def grid(META):
        return (
            sum(triton.cdiv(m, META["BLOCK_M"]) for m in ms)
            * sum(triton.cdiv(n, META["BLOCK_N"]) for n in ns),
            META["SPLIT_K"],
        )

    _xformers_tiled_matmul_kernel[grid](
        *[
            a[min(i, len(a) - 1)][min(j, len(a[0]) - 1)]
            for i in range(3)
            for j in range(3)
        ],
        *[
            b[min(i, len(b) - 1)][min(j, len(b[0]) - 1)]
            for i in range(3)
            for j in range(3)
        ],
        *[
            c[min(i, len(c) - 1)][min(j, len(c[0]) - 1)]
            for i in range(3)
            for j in range(3)
        ],
        *[ms[i] if len(ms) > i else 0 for i in range(3)],
        *[ns[i] if len(ns) > i else 0 for i in range(3)],
        *[ks[i] if len(ks) > i else 0 for i in range(3)],
        *strides_am,
        *strides_ak,
        *strides_bk,
        *strides_bn,
        *strides_cm,
        *strides_cn,
        ACC_TYPE=ACC_TYPE,
    )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import hashlib
import math
from typing import Callable, List, Optional, Tuple, Type

import pytest
import torch
from xformers.ops import fmha
from xformers.ops.fmha.common import AttentionFwOpBase
from xformers.ops.fmha.merge_training import (
    memory_efficient_attention_partial_autograd,
    merge_attentions_autograd,
    Partial,
)

from .utils import assert_allclose, disable_on_rocm

compute_capability = (0, 0)
if torch.cuda.is_available():
    compute_capability = torch.cuda.get_device_capability("cuda")
sm80_or_better_only = pytest.mark.skipif(
    compute_capability < (8, 0), reason="requires sm90+"
)
sm90_or_better_only = pytest.mark.skipif(
    compute_capability < (9, 0), reason="requires sm90+"
)


# This temporary working is necessary because the MTIA test collection might not happen
# on the same device as the device the tests are actually executed on. If test collection
# is done on a device without MTIA, the supported masks will contain masks that MTIA support
# and the corresponding tests will get collected. But when it comes time to actually run the
# tests, the mask won't be supported because it is run on an actual MTIA device.
def get_supported_attn_bias_types(op):
    supported_attn_bias_types = op.SUPPORTED_ATTN_BIAS_TYPES

    try:
        import mtia.host_runtime.torch_mtia.dynamic_library  # noqa

        supported_attn_bias_types = [
            b
            for b in supported_attn_bias_types
            if not issubclass(
                b,
                (
                    fmha.attn_bias.PagedBlockDiagonalGappyKeysMask,
                    fmha.attn_bias.PagedBlockDiagonalPaddedKeysMask,
                ),
            )
        ]
    except (ImportError, OSError):
        pass

    return supported_attn_bias_types


@disable_on_rocm
@sm80_or_better_only
@pytest.mark.parametrize(
    "op",
    [
        fmha.triton_splitk.FwOp,
        fmha.flash.FwOp,
        fmha.flash3.FwOp,
        None,
    ],
    ids=lambda op: "None" if op is None else op.NAME,
)
@pytest.mark.parametrize("G,H", [(1, 11), (7, 1), (1, 1), (7, 11), (None, 11)])
@pytest.mark.parametrize(
    "write_lse", (False, True), ids=lambda x: "write_lse" if x else ""
)
@pytest.mark.parametrize(
    "stack_inputs", (False, True), ids=lambda x: "stack_inputs" if x else ""
)
def test_merge_attentions_nobias(
    write_lse: bool,
    stack_inputs: bool,
    op: Type[AttentionFwOpBase],
    G: Optional[int],
    H: int,
):
    """
    Merging the same attention twice shouldn't change anything.
    This also tests the shape of the lse output of each permitted op.
    """
    if op is fmha.flash3.FwOp and not op.is_available():
        pytest.skip("Flash3 not available")
    B, Mq, K = 13, 3, 192
    if op is fmha.triton_splitk.FwOp:
        K = 128
    case_name = str((write_lse, G, H, stack_inputs)).encode("ascii")
    many_keys = hashlib.md5(case_name).digest()[0] % 2
    M = [5, 100000][many_keys]
    if op is None or torch.bfloat16 in op.SUPPORTED_DTYPES:
        dtype = torch.bfloat16
    else:
        dtype = next(iter(op.SUPPORTED_DTYPES))
    if dtype == torch.float8_e4m3fn:
        pytest.skip("float8 not supported")
    if G is None:
        q = 3 * torch.rand(B, Mq, H, K, dtype=dtype, device="cuda")
        k = (3 * torch.rand(B, M, 1, K, dtype=dtype, device="cuda")).expand(B, M, H, K)
        v = (3 * torch.rand(B, M, 1, K, dtype=dtype, device="cuda")).expand(B, M, H, K)
    else:
        q = 3 * torch.rand(B, Mq, G, H, K, dtype=dtype, device="cuda")
        k = (3 * torch.rand(B, M, G, 1, K, dtype=dtype, device="cuda")).expand(
            B, M, G, H, K
        )
        v = (3 * torch.rand(B, M, G, 1, K, dtype=dtype, device="cuda")).expand(
            B, M, G, H, K
        )
    out1, lse1 = fmha.memory_efficient_attention_partial(q, k, v, op=op)
    assert out1.shape == q.shape
    M_ceil = lse1.shape[-1]
    assert M_ceil >= Mq
    assert lse1.shape == (B, H, M_ceil) if G is None else (B, G, H, M_ceil)
    lse1 = lse1[..., :Mq]

    attn_chunks = [out1, out1]
    lse_chunks = [lse1, lse1]
    attn_chunks_ = torch.stack(attn_chunks) if stack_inputs else attn_chunks
    lse_chunks_ = torch.stack(lse_chunks) if stack_inputs else lse_chunks
    out, lse = fmha.merge_attentions(attn_chunks_, lse_chunks_, write_lse=write_lse)  # type: ignore
    assert out.shape == out1.shape
    assert_allclose(out1, out, rtol=1e-3, atol=1e-3, msg="out")
    if write_lse:
        assert lse is not None
        assert lse.shape[:-1] == lse1.shape[:-1]
        assert_allclose(
            lse1[..., :Mq] + math.log(2), lse[..., :Mq], rtol=1e-3, atol=1e-3, msg="lse"
        )
    else:
        assert lse is None


@disable_on_rocm
@sm80_or_better_only
@pytest.mark.parametrize(
    "dtype,op",
    [
        (torch.bfloat16, fmha.triton_splitk.FwOp_S1),
        # Cutlass's LSE is not consistent
        # (torch.float32, fmha.cutlass.FwOp),
        (torch.bfloat16, fmha.flash.FwOp),
    ],
    ids=lambda o: f"{o.NAME}" if hasattr(o, "NAME") else str(o),
)
@pytest.mark.parametrize("num_queries", [1])
@pytest.mark.parametrize("bmghk", [True, False], ids=lambda x: "bmghk" if x else "")
def test_partial_paged(
    dtype: torch.dtype, op: Type[AttentionFwOpBase], num_queries: int, bmghk: bool
):
    B = 128
    N_H_L = 8
    D_H = 128
    page_size = 256
    G = 2 if bmghk else 1
    block_tables = torch.zeros((B, 1), dtype=torch.int32, device="cuda")
    torch.manual_seed(1)
    output_dtype = torch.float32 if op.SUPPORTS_OUTPUT_DTYPE else None

    B_T = num_queries * B

    q = torch.randn((1, B_T, G, N_H_L, D_H), dtype=dtype, device="cuda")
    k = torch.randn((1, page_size, G, 1, D_H), dtype=dtype, device="cuda")
    v = torch.randn_like(k)
    k = k.expand(1, page_size, G, N_H_L, D_H)
    v = v.expand(1, page_size, G, N_H_L, D_H)
    if not bmghk:
        q = q[:, :, 0]
        k = k[:, :, 0]
        v = v[:, :, 0]

    attn_bias = (
        fmha.attn_bias.PagedBlockDiagonalCausalWithOffsetPaddedKeysMask.from_seqlens(
            q_seqlen=[num_queries] * B,
            kv_seqlen=[1] + ([100] * (B - 1)),
            page_size=page_size,
            block_tables=block_tables,
        )
    )

    if attn_bias not in get_supported_attn_bias_types(op):
        pytest.skip("Not supported bias")

    attn_chunk, lse_chunk = fmha.memory_efficient_attention_partial(
        q,
        k,
        v,
        attn_bias,
        op=op,
        output_dtype=output_dtype,
    )
    if bmghk:
        assert attn_chunk.shape == (1, B_T, G, N_H_L, D_H)
        assert lse_chunk.shape == (
            1,
            G,
            N_H_L,
            B_T,
        ), f"{lse_chunk.shape=}, {(1, G, N_H_L, B_T)=}"
    else:
        assert attn_chunk.shape == (1, B_T, N_H_L, D_H)
        assert lse_chunk.shape == (
            1,
            N_H_L,
            B_T,
        ), f"{lse_chunk.shape=}, {(1, N_H_L, B_T)=}"


@disable_on_rocm
@sm80_or_better_only
@pytest.mark.parametrize(
    "dtype,op",
    [
        (torch.bfloat16, fmha.triton_splitk.FwOp_S1),
        (torch.bfloat16, fmha.triton_splitk.FwOp_S32),
        # Cutlass's LSE is not consistent
        # (torch.float32, fmha.cutlass.FwOp),
        (torch.bfloat16, fmha.flash.FwOp),
    ],
    ids=lambda o: f"{o.NAME}" if hasattr(o, "NAME") else str(o),
)
@pytest.mark.parametrize("num_queries", [1, 2])
@pytest.mark.parametrize("bmghk", [True, False], ids=lambda x: "bmghk" if x else "")
@pytest.mark.parametrize(
    "stack_inputs", (False, True), ids=lambda x: "stack_inputs" if x else ""
)
def test_merge_attentions_decoding(
    dtype: torch.dtype,
    op: Type[AttentionFwOpBase],
    num_queries: int,
    bmghk: bool,
    stack_inputs: bool,
):
    """
    Compute decoding attention on chunks of K/V and merge them together.
    Compare with computing attention on the whole K/V.
    """
    MAX_T = 8192
    B = 128
    N_H_L = 8
    D_H = 128
    G = 2 if bmghk else 1
    torch.manual_seed(1)
    output_dtype = torch.float32 if op.SUPPORTS_OUTPUT_DTYPE else None

    num_chunks = 10

    chunk_starts = sorted(
        torch.randint(low=1, high=MAX_T // 2, size=(num_chunks,)).tolist()
    )
    chunk_starts[0] = 0
    chunk_starts.append(MAX_T)

    # We construct sequences so that even the last chunk has a non-empty part of every sequence
    # as long as the number of queries.
    # Otherwise the corresponding LSE will be -inf and that'll propagate to the whole sum.
    # It is possible to teach the kernel to ignore infinite LSEs, but in practical use cases
    # of merging attention, e.g. a batch of sequences with a common prefix, this condition should be satisfied.
    k_lens = torch.randint(
        low=chunk_starts[-2] + num_queries, high=MAX_T, size=(B,)
    ).tolist()
    q_lens = [num_queries] * B
    B_T = num_queries * B

    q = torch.randn((1, B_T, G, N_H_L, D_H), dtype=dtype, device="cuda")
    k = torch.randn((B, MAX_T, G, 1, D_H), dtype=dtype, device="cuda")
    v = torch.randn_like(k)
    if not bmghk:
        q = q[:, :, 0]

    # Compute per-chunk attention
    chunks_output: List[Tuple[torch.Tensor, torch.Tensor]] = []
    for i in range(num_chunks):
        chunk_start, chunk_end = chunk_starts[i], chunk_starts[i + 1]
        k_chunk = k[:, chunk_start:chunk_end, ...]
        v_chunk = v[:, chunk_start:chunk_end, ...]
        axk = k_chunk.reshape(-1, G, 1, D_H).expand(1, -1, G, N_H_L, D_H)
        axv = v_chunk.reshape(-1, G, 1, D_H).expand(1, -1, G, N_H_L, D_H)
        if not bmghk:
            axk = axk[:, :, 0]
            axv = axv[:, :, 0]

        bias_type = fmha.attn_bias.BlockDiagonalPaddedKeysMask
        if i + 1 == num_chunks:
            bias_type = fmha.attn_bias.BlockDiagonalCausalWithOffsetPaddedKeysMask
        attn_bias = bias_type.from_seqlens(
            q_seqlen=q_lens,
            kv_padding=chunk_end - chunk_start,
            kv_seqlen=[max(min(x, chunk_end) - chunk_start, 0) for x in k_lens],
        )

        attn_chunk, lse_chunk = fmha.memory_efficient_attention_partial(
            q,
            axk,
            axv,
            attn_bias,
            op=op,
            output_dtype=output_dtype,
        )
        if bmghk:
            assert attn_chunk.shape == (1, B_T, G, N_H_L, D_H)
            assert lse_chunk.shape == (1, G, N_H_L, B_T)
        else:
            assert attn_chunk.shape == (1, B_T, N_H_L, D_H)
            assert lse_chunk.shape == (1, N_H_L, B_T)
        chunks_output.append((attn_chunk, lse_chunk))

    # Merge attention from all chunks
    attn_split = [attn_chunk for attn_chunk, _ in chunks_output]
    lse_split = [lse_chunk for _, lse_chunk in chunks_output]
    if stack_inputs:
        attn_out, lse_out = fmha.merge_attentions(
            torch.stack(attn_split), torch.stack(lse_split), output_dtype=dtype
        )
    else:
        attn_out, lse_out = fmha.merge_attentions(
            attn_split, lse_split, output_dtype=dtype
        )
    assert lse_out is not None

    # Compute attention on the full K/V
    attn_bias = fmha.attn_bias.BlockDiagonalCausalWithOffsetPaddedKeysMask.from_seqlens(
        q_seqlen=q_lens,
        kv_padding=MAX_T,
        kv_seqlen=k_lens,
    )
    axk = k.view(1, -1, G, 1, D_H).expand(1, -1, G, N_H_L, D_H)
    axv = v.view(1, -1, G, 1, D_H).expand(1, -1, G, N_H_L, D_H)
    if not bmghk:
        axk = axk[:, :, 0]
        axv = axv[:, :, 0]
    attn_full, lse_full = fmha.memory_efficient_attention_partial(
        q,
        axk,
        axv,
        attn_bias,
        op=op,
        output_dtype=output_dtype,
    )

    assert_allclose(
        lse_out.to(lse_full.dtype), lse_full, rtol=1e-3, atol=1e-3, msg="lse"
    )
    assert_allclose(
        attn_out.to(attn_full.dtype), attn_full, rtol=1e-3, atol=1e-3, msg="out"
    )

    attn_full2 = fmha.memory_efficient_attention_forward(
        q,
        axk,
        axv,
        attn_bias,
        op=op,
        output_dtype=output_dtype,
    )
    assert_allclose(attn_full2, attn_full, rtol=1e-3, atol=1e-3, msg="out2")


@disable_on_rocm
@sm80_or_better_only
@pytest.mark.parametrize(
    "dtype,op",
    [
        (torch.bfloat16, fmha.triton_splitk.FwOp_S1),
        (torch.bfloat16, fmha.triton_splitk.FwOp_S32),
    ],
    ids=lambda o: f"{o.NAME}" if hasattr(o, "NAME") else str(o),
)
@pytest.mark.parametrize("gqa", [False, True], ids=lambda x: "gqa" if x else "")
def test_merge_attentions_sharedinput(
    dtype: torch.dtype,
    op: Type[AttentionFwOpBase],
    gqa: bool,
):
    """
    Compute decoding attention on chunks of K/V and merge them together.
    Compare with computing attention on the whole K/V.
    """
    MAX_T = 8192
    N_H_L = 16
    D_H = 128
    G = 2
    torch.manual_seed(1)
    output_dtype = torch.float32 if op.SUPPORTS_OUTPUT_DTYPE else None

    shared_length = 20
    full_lengths = [30, 35, 40]

    attn_bias = fmha.attn_bias.BlockDiagonalCausalWithOffsetPaddedKeysMask.from_seqlens(
        q_seqlen=[1, 1, 1],
        kv_padding=MAX_T,
        kv_seqlen=full_lengths,
    )
    attn_bias1 = fmha.attn_bias.BlockDiagonalPaddedKeysMask.from_seqlens(
        q_seqlen=[2, 1],
        kv_padding=MAX_T,
        kv_seqlen=[shared_length, 0],
    )
    attn_bias2 = fmha.attn_bias.BlockDiagonalGappyKeysMask.from_seqlens(
        q_seqlen=[1, 1, 1],
        kv_seqstarts=[shared_length, MAX_T + shared_length, 2 * MAX_T, 3 * MAX_T],
        kv_seqlen=[
            full_lengths[0] - shared_length,
            full_lengths[1] - shared_length,
            full_lengths[2],
        ],
    )

    q = torch.randn((1, 3, G, N_H_L, D_H), dtype=dtype, device="cuda")
    k = torch.randn((3, MAX_T, G, 1 if gqa else N_H_L, D_H), dtype=dtype, device="cuda")
    v = torch.randn_like(k)
    k[1, :shared_length] = k[0, :shared_length]
    v[1, :shared_length] = v[0, :shared_length]
    k = k.flatten(end_dim=1)[None]
    v = v.flatten(end_dim=1)[None]
    k = k.expand((1, 3 * MAX_T, G, N_H_L, D_H))
    v = v.expand((1, 3 * MAX_T, G, N_H_L, D_H))

    attn_chunk1, lse_chunk1 = fmha.memory_efficient_attention_partial(
        q,
        k,
        v,
        attn_bias1,
        op=op,
        output_dtype=output_dtype,
    )
    assert attn_chunk1.shape == (1, 3, G, N_H_L, D_H)
    assert lse_chunk1.shape == (1, G, N_H_L, 3)
    if gqa:
        attn_chunk1a, lse_chunk1a = fmha.memory_efficient_attention_partial(
            q,
            k.contiguous(),
            v,
            attn_bias1,
            op=op,
            output_dtype=output_dtype,
        )
        assert attn_chunk1a.shape == (1, 3, G, N_H_L, D_H)
        assert lse_chunk1a.shape == (1, G, N_H_L, 3)
        assert_allclose(
            attn_chunk1a.nan_to_num(0, 0, 0), attn_chunk1.nan_to_num(0, 0, 0)
        )
        assert_allclose(lse_chunk1a.nan_to_num(0, 0, 0), lse_chunk1.nan_to_num(0, 0, 0))

    attn_chunk2, lse_chunk2 = fmha.memory_efficient_attention_partial(
        q,
        k,
        v,
        attn_bias2,
        op=op,
        output_dtype=output_dtype,
    )
    assert attn_chunk2.shape == (1, 3, G, N_H_L, D_H)
    assert lse_chunk2.shape == (1, G, N_H_L, 3)
    # Merge attention from all chunks

    attn_out, lse_out = fmha.merge_attentions(
        [attn_chunk1, attn_chunk2],
        [lse_chunk1, lse_chunk2],
        output_dtype=dtype,  # type: ignore
    )
    assert lse_out is not None

    # Compute attention on the full K/V
    attn_full, lse_full = fmha.memory_efficient_attention_partial(
        q,
        k,
        v,
        attn_bias,
        op=op,
        output_dtype=output_dtype,
    )
    assert_allclose(
        attn_out.to(attn_full.dtype), attn_full, rtol=1e-2, atol=2e-3, msg="out"
    )
    assert_allclose(
        lse_out.to(lse_full.dtype), lse_full, rtol=1e-3, atol=1e-3, msg="lse"
    )


@sm80_or_better_only
@pytest.mark.parametrize("bmghk", (False, True))
def test_merge_attentions_against_ref(bmghk: bool):
    split_k = 16
    B = 12
    M = 137
    G = 2 if bmghk else 1
    N_H_L = 8
    D_H = 128
    dtype = torch.float32

    attn_split = torch.randn([split_k, B, M, G, N_H_L, D_H], dtype=dtype, device="cuda")
    lse_split = torch.randn([split_k, B, G, N_H_L, M], dtype=dtype, device="cuda")

    if not bmghk:
        attn_split = attn_split[:, :, :, 0]
        lse_split = lse_split[:, :, 0]

    attn_out_ref, lse_out_ref = _merge_attentions_ref(attn_split, lse_split)
    attn_out, lse_out = fmha.merge_attentions(attn_split, lse_split)

    torch.testing.assert_close(lse_out, lse_out_ref, rtol=1e-4, atol=1e-4)
    torch.testing.assert_close(attn_out, attn_out_ref, rtol=1e-4, atol=1e-4)


def _merge_attentions_ref(attn_split, lse_split):
    """
    attn_split: [split_k, B, M, (G,) H, Kq]
    lse_split: [split_k, B, (G,) H, M]
    """
    is_bmghk = len(attn_split.shape) == 6
    if not is_bmghk:
        attn_split = attn_split.unsqueeze(3)
        lse_split = lse_split.unsqueeze(2)

    lse_split = lse_split[..., None].moveaxis(4, 2)  # [split_k, B, M, G, H, 1]

    lse_max, _ = torch.max(lse_split, dim=0)  # [B, M, G, H, 1]
    sumexp_normalized = torch.exp(lse_split - lse_max)  # [split_k, B, M, G, H, 1]
    denominator = sumexp_normalized.sum(dim=0)  # [B, M, G, H, 1]
    numerator = (sumexp_normalized * attn_split).sum(dim=0)  # [B, M, G, H, K]

    attn_out = numerator / denominator  # [B, M_ceil, G, H, Kq]
    lse_out = lse_max + torch.log(denominator)
    lse_out = lse_out.squeeze(4).permute(0, 2, 3, 1)  # [B, G, H, M]

    if not is_bmghk:
        attn_out = attn_out.squeeze(2)
        lse_out = lse_out.squeeze(1)

    return attn_out, lse_out


@sm80_or_better_only
def test_merge_attention_with_compile() -> None:
    op = fmha.flash3.FwOp
    if not op.is_available():
        pytest.skip("Op is not available")
    dtype = torch.bfloat16
    B, M, H, K = 1, 256, 2, 128
    q, k, v = [
        (3 * torch.rand(B, M, H, K, dtype=dtype, device="cuda")) for _ in range(3)
    ]

    out1, lse1 = fmha.memory_efficient_attention_partial(q, k, v, op=op)

    def run_code() -> torch.Tensor:
        out, _ = fmha.merge_attentions([out1], [lse1], write_lse=True)
        return out

    out_ref = run_code()
    out_c = torch.compile(run_code, fullgraph=True)()

    assert torch.allclose(out_ref, out_c, atol=1e-2, rtol=1e-2)

    q.requires_grad_(True)
    out1, lse1 = fmha.memory_efficient_attention_partial(q, k, v, op=op)
    loss = fmha.merge_attentions([out1], [lse1])[0].sum()
    with pytest.raises(
        NotImplementedError,
        match="Backward pass is not implemented for merge_attentions",
    ):
        loss.backward()


@sm80_or_better_only
def test_merge_training():
    torch.manual_seed(1)
    B, M, H, K = 1, 50, 1, 128
    dtype = torch.bfloat16
    op = (fmha.flash3.FwOp, fmha.flash3.BwOp)
    q = 3 * torch.rand((B, M, H, K), device="cuda", dtype=dtype)
    k = 3 * torch.rand((B, M, H, K), device="cuda", dtype=dtype)
    v = 3 * torch.rand((B, M, H, K), device="cuda", dtype=dtype)
    grad_out = 3 * torch.rand((B, M, H, K), device="cuda", dtype=dtype)

    total_attention, (q_grad, k_grad, v_grad) = torch.autograd.functional.vjp(
        fmha.memory_efficient_attention, (q, k, v), grad_out
    )

    def total_attn_via_Partial(q_, k_, v_):
        return merge_attentions_autograd(
            memory_efficient_attention_partial_autograd(q_, k_, v_)
        )

    attn, (q_grad_, k_grad_, v_grad_) = torch.autograd.functional.vjp(
        total_attn_via_Partial, (q, k, v), grad_out
    )

    assert_allclose(attn, total_attention, rtol=1e-1, atol=1e-3, msg="out")
    assert_allclose(k_grad_, k_grad, rtol=1e-2, atol=1e-3, msg="dk_")
    assert_allclose(q_grad_, q_grad, rtol=1e-2, atol=1e-3, msg="dq_")
    assert_allclose(v_grad_, v_grad, rtol=1e-2, atol=1e-3, msg="dv_")

    def attn_via_Partial(q_, k_, v_):
        split = M // 2
        k1 = k_[:, :split]
        k2 = k_[:, split:]
        v1 = v_[:, :split]
        v2 = v_[:, split:]

        partial1 = memory_efficient_attention_partial_autograd(q_, k1, v1, op=op)
        partial2 = memory_efficient_attention_partial_autograd(q_, k2, v2, op=op)
        return merge_attentions_autograd(partial1, partial2)

    merged, (q_grad_, k_grad_, v_grad_) = torch.autograd.functional.vjp(
        attn_via_Partial, (q, k, v), grad_out
    )
    assert_allclose(
        merged.to(total_attention.dtype),
        total_attention,
        rtol=1e-1,
        atol=1e-3,
        msg="out",
    )
    assert_allclose(k_grad_, k_grad, rtol=0.1, atol=0.9, msg="dk")
    assert_allclose(q_grad_, q_grad, rtol=0.1, atol=0.9, msg="dq")
    assert_allclose(v_grad_, v_grad, rtol=0.1, atol=0.9, msg="dv")


def _pad_seqdim(partial: Partial, left: int, right: int) -> Partial:
    padding = (0, 0) * (3 if partial.is_bmghk() else 2) + (left, right)
    return partial.apply(lambda x: torch.nn.functional.pad(x, padding))


def _slice(partial: Partial, a: int, b: int) -> Partial:
    return partial.apply(lambda x: x[:, a:b])


@sm80_or_better_only
def test_merge_training_compile():
    torch.manual_seed(1)
    B, M, H, K = 1, 50, 1, 128
    dtype = torch.bfloat16
    q = 3 * torch.rand((B, M, H, K), device="cuda", dtype=dtype)
    k = 3 * torch.rand((B, M, H, K), device="cuda", dtype=dtype)
    v = 3 * torch.rand((B, M, H, K), device="cuda", dtype=dtype)

    def f(
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
    ) -> torch.Tensor:
        k1 = k[:, : M // 2]
        k2 = k[:, M // 2 :]
        v1 = v[:, : M // 2]
        v2 = v[:, M // 2 :]
        partial1 = memory_efficient_attention_partial_autograd(q, k1, v1)
        partial2 = memory_efficient_attention_partial_autograd(q, k2, v2)
        partial2 = _pad_seqdim(_pad_seqdim(partial2, 2, 3), -2, -3)
        partial2 = _slice(_pad_seqdim(partial2, 2, 3), 2, -3)
        merged = merge_attentions_autograd(partial1, partial2)
        return merged.sum()

    out, grads = torch.autograd.functional.vjp(f, (q, k, v))
    outc, gradsc = torch.autograd.functional.vjp(
        torch.compile(fullgraph=True)(f), (q, k, v)
    )
    assert_allclose(out, outc, rtol=1e-1, atol=1e-3)
    for i, (grad, gradc) in enumerate(zip(grads, gradsc)):
        assert_allclose(grad, gradc, rtol=1e-1, atol=1e-3, msg=f"grad{i}")

    def g(q: torch.Tensor, k1: torch.Tensor, v1: torch.Tensor) -> torch.Tensor:
        partial1 = memory_efficient_attention_partial_autograd(q, k1, v1)
        return merge_attentions_autograd(partial1).sum()

    out, grads = torch.autograd.functional.vjp(g, (q, k, v))
    outc, gradsc = torch.autograd.functional.vjp(
        torch.compile(fullgraph=True)(g), (q, k, v)
    )
    assert_allclose(out, outc, rtol=1e-1, atol=1e-3)
    for i, (grad, gradc) in enumerate(zip(grads, gradsc)):
        assert_allclose(grad, gradc, rtol=1e-1, atol=1e-3, msg=f"grad{i}")


@sm80_or_better_only
def test_merge_training_zilch():
    with pytest.raises(ValueError, match="No partials to merge"):
        merge_attentions_autograd()


@sm80_or_better_only
def test_merge_training_undilate():
    torch.manual_seed(1)

    def undilate(factor: int) -> Callable[[torch.Tensor], torch.Tensor]:
        """
        For a given factor, operate on BMHK attention output as follows:
        If sequence length is M and there are H * factor heads,
        redistribute so there are H heads and each
        original sequence position is factor times inflated.
        """

        def inner(x: torch.Tensor) -> torch.Tensor:
            M = x.shape[1]
            H = x.shape[2] // factor
            return x.flatten(1, 2).unflatten(1, (M * factor, H))

        return inner

    B, M, F, H, K = 1, 2, 3, 5, 128
    dtype = torch.bfloat16
    q = 3 * torch.rand((B, M, F * H, K), device="cuda", dtype=dtype)
    k = 3 * torch.rand((B, M, F * H, K), device="cuda", dtype=dtype)
    v = 3 * torch.rand((B, M, F * H, K), device="cuda", dtype=dtype)

    bias = fmha.BlockDiagonalMask.from_seqlens([1] * M)
    partial = memory_efficient_attention_partial_autograd(q, k, v, bias)

    q = q.reshape(B, M * F, H, K)
    k = k.reshape(B, M * F, H, K)
    v = v.reshape(B, M * F, H, K)

    bias = fmha.BlockDiagonalMask.from_seqlens([1] * (M * F))
    expected = memory_efficient_attention_partial_autograd(q, k, v, bias)
    undilated = partial.apply(undilate(F))

    assert_allclose(undilated._attn, expected._attn)
    assert_allclose(undilated._lse, expected._lse)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import torch
import triton
import triton.language as tl


@triton.jit
def index_select_cat_fwd_kernel(
    output_ptr,  # *Pointer* to output tensor.
    source_ptr,  # *Pointer* to source tensor.
    index_ptr,  # *Pointer* to index tensor.
    num_indices,
    num_cols,
    stride0,  # Stride information of source tensor.
    stride1,
    BLOCK_SIZE_INDEX: tl.constexpr,  # Number of indices each program should process.
    BLOCK_SIZE_COL: tl.constexpr,  # Number of cols each program should process.
):
    pid0 = tl.program_id(axis=0)  # We use 2D launch grid
    pid1 = tl.program_id(axis=1)

    indices = pid0 * BLOCK_SIZE_INDEX + tl.arange(0, BLOCK_SIZE_INDEX)
    rows = tl.load(index_ptr + indices, mask=(indices < num_indices))
    cols = pid1 * BLOCK_SIZE_COL + tl.arange(0, BLOCK_SIZE_COL)

    source_offsets = source_ptr + rows[:, None] * stride0 + cols[None, :] * stride1
    mask = (indices[:, None] < num_indices) & (cols[None, :] < num_cols)
    output = tl.load(source_offsets, mask=mask)

    output_offsets = output_ptr + indices[:, None] * stride0 + cols[None, :] * stride1
    tl.store(output_offsets, output, mask=mask)


def index_select_cat_fwd(
    output: torch.Tensor,
    source: torch.Tensor,
    index: torch.Tensor,
):
    if not (source.is_cuda and index.is_cuda):
        raise ValueError("The index tensor and the source tensor must be of type CUDA!")

    if not source.ndim == 2:
        raise ValueError(f"Expected 2-dimensional tensor, got {source.ndim}.")
    if not index.ndim == 1:
        raise ValueError(f"Expected 1-dimensional tensor, got {index.ndim}.")

    num_rows, num_cols = source.shape
    num_indices = index.shape[0]

    if not num_indices < num_rows:
        raise ValueError(
            "The number of indices cannot exceed the number of rows in the source matrix."
        )

    stride0, stride1 = source.stride(0), source.stride(1)

    def grid(meta):
        return (
            triton.cdiv(num_indices, meta["BLOCK_SIZE_INDEX"]),
            triton.cdiv(num_cols, meta["BLOCK_SIZE_COL"]),
        )

    index_select_cat_fwd_kernel[grid](
        output,
        source,
        index,
        num_indices,
        num_cols,
        stride0,
        stride1,
        BLOCK_SIZE_INDEX=1,
        BLOCK_SIZE_COL=512,
    )

    return output


@triton.jit
def index_select_cat_bwd_kernel(
    grad_source_ptr,  # *Pointer* to grad_source tensor.
    index_ptr,  # *Pointer* to index tensor.
    grad_output_ptr,  # *Pointer* to grad_output tensor.
    num_rows,
    num_indices,
    num_cols,
    stride0,  # Stride information of input and source tensor.
    stride1,
    BLOCK_SIZE_INDEX: tl.constexpr,  # Number of indices each program should process.
    BLOCK_SIZE_COL: tl.constexpr,  # Number of cols each program should process.
):
    pid0 = tl.program_id(axis=0)  # We use 3D launch grid
    pid1 = tl.program_id(axis=1)

    cols = pid1 * BLOCK_SIZE_COL + tl.arange(0, BLOCK_SIZE_COL)

    # load grad_output
    grad_output_indices = pid0 * BLOCK_SIZE_INDEX + tl.arange(0, BLOCK_SIZE_INDEX)
    grad_output_offsets = (
        grad_output_ptr
        + grad_output_indices[:, None] * stride0
        + cols[None, :] * stride1
    )
    grad_output_mask = (grad_output_indices[:, None] < num_indices) & (
        cols[None, :] < num_cols
    )
    grad_output = tl.load(grad_output_offsets, mask=grad_output_mask).to(tl.float32)

    # select indices from grad_source
    grad_source_indices = tl.load(
        index_ptr + grad_output_indices, mask=(grad_output_indices < num_indices)
    )
    grad_source_offsets = (
        grad_source_ptr
        + grad_source_indices[:, None] * stride0
        + cols[None, :] * stride1
    )

    # compute scaled index add and save
    tl.store(grad_source_offsets, grad_output, mask=grad_output_mask)


def index_select_cat_bwd(
    grad_source: torch.Tensor,
    index: torch.Tensor,
    grad_output: torch.Tensor,
):
    if not (grad_source.is_cuda and grad_output.is_cuda):
        raise ValueError("The grad_source and grad_output tensor must be of type CUDA!")

    if not (grad_source.ndim == 2 and grad_output.ndim == 2):
        raise ValueError(
            f"The grad_source and grad_output must be three-dimensional "
            f"(got {grad_source.ndim} and {grad_output.ndim})!"
        )
    if not grad_source.shape[1] == grad_output.shape[1]:
        raise ValueError(
            f"The number of elements along dimension 1 of grad_source and grad_output must be the same "
            f"(got {grad_source.shape[1]} and {grad_output.shape[1]})"
        )

    num_rows, num_cols = grad_source.shape
    num_indices, num_cols = grad_output.shape
    if not num_rows >= num_indices:
        raise ValueError(
            f"The number of elements along dimension 0 of grad_source must be larger than that of grad_output "
            f"(got {num_rows} and {num_indices})!"
        )
    if not index.shape[0] == num_indices:
        raise ValueError(
            f"The number of indices and the number of elements along dimension 0 of grad_output must match "
            f"(got {index.shape[0]} and {num_indices})!"
        )

    stride0, stride1 = grad_source.stride(0), grad_source.stride(1)
    if not (grad_output.stride(0) == stride0 and grad_output.stride(1) == stride1):
        raise ValueError(
            f"The strides of the grad_source and grad_output tensors must match "
            f"(got {stride0} vs. {grad_output.stride(0)}, {stride1} vs. {grad_output.stride(1)})!"
        )

    def grid(meta):
        return (
            triton.cdiv(num_indices, meta["BLOCK_SIZE_INDEX"]),
            triton.cdiv(num_cols, meta["BLOCK_SIZE_COL"]),
        )

    index_select_cat_bwd_kernel[grid](
        grad_source,
        index,
        grad_output,
        num_rows,
        num_indices,
        num_cols,
        grad_source.stride(0),
        grad_source.stride(1),
        BLOCK_SIZE_INDEX=1,
        BLOCK_SIZE_COL=512,
    )

    return



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.
# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.
from typing import Callable, Optional, Tuple, Type, Union

import torch

from xformers.ops import fmha


"""
Friendly wrapper around merge_attentions which works with autograd.

Use as follows

```
partial1 = memory_efficient_attention_partial_autograd(q, k1, v1, ...)
partial2 = memory_efficient_attention_partial_autograd(q, k2, v2, ...)
attn_out = merge_attentions_autograd(partial1, partial2)
```

merge_attentions_autograd() can take any number of inputs. Note that
partial1 and partial2 are not tensors, but rather objects of type
`Partial`.

If you have partial1 and you changed your mind and don't
want to merge it with anything, you can do
```
attn_out = merge_attentions_autograd(partial1)
```

"""


class _PartialFunc(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx: torch.autograd.function.FunctionCtx,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        attn_bias: Optional[Union[torch.Tensor, fmha.AttentionBias]],
        p: float = 0.0,
        scale: Optional[float] = None,
        op: Optional[Union[fmha.AttentionOp, Type[fmha.AttentionFwOpBase]]] = None,
        output_dtype: Optional[torch.dtype] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        ctx.bias = attn_bias  # type: ignore
        ctx.save_for_backward(query, key, value)
        ctx.p = p  # type: ignore
        ctx.scale = scale  # type: ignore
        ctx.op = op[1] if isinstance(op, tuple) else None  # type: ignore
        attn, lse = fmha.memory_efficient_attention_partial(
            query, key, value, attn_bias, p, scale, op=op, output_dtype=output_dtype
        )
        placeholder = torch.empty_like(attn)
        return attn, lse, placeholder

    @staticmethod
    def backward(  # type: ignore[override]
        ctx: torch.autograd.function.FunctionCtx,
        grad_attn: torch.Tensor,
        lse: torch.Tensor,
        out: torch.Tensor,
    ) -> Tuple[Optional[torch.Tensor], ...]:
        query, key, value = ctx.saved_tensors  # type: ignore
        grad_q, grad_k, grad_v = fmha.memory_efficient_attention_backward(
            grad_attn,
            out,
            lse.contiguous(),
            query,
            key,
            value,
            ctx.bias,  # type: ignore
            ctx.p,  # type: ignore
            ctx.scale,  # type: ignore
            op=ctx.op,  # type: ignore
        )
        return grad_q, grad_k, grad_v, None, None, None, None, None


class _MergeFunc(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx: torch.autograd.function.FunctionCtx,
        *inputs: torch.Tensor,
    ) -> torch.Tensor:
        ctx.length = len(inputs) // 3  # type: ignore
        if ctx.length > 1:  # type: ignore
            attns = inputs[::3]
            lses = inputs[1::3]
            out, lse = fmha.merge_attentions(attns, lses)
            assert lse is not None
        else:
            out, lse = inputs[:2]
        ctx.save_for_backward(out, lse)
        return out

    @staticmethod
    def backward(  # type: ignore[override]
        ctx: torch.autograd.function.FunctionCtx, grad_out: torch.Tensor
    ) -> Tuple[Optional[torch.Tensor], ...]:
        out, lse = ctx.saved_tensors  # type: ignore
        return (grad_out, lse, out) * ctx.length  # type: ignore


class Partial:
    """
    This class is used to represent a partial attention output, which is
    returned by `memory_efficient_attention_partial_autograd`.

    Attributes: (Do not access them directly, use the methods instead.)

        _attn: torch.Tensor
        _lse: torch.Tensor . (Its grad is the full LSE to be used by
            the individual backward passes.)
        _placeholder: torch.Tensor, whose grad is used for passing the full attention
            output to the individual backward passes.
    """

    def __init__(
        self,
        attn: torch.Tensor,
        lse: torch.Tensor,
        placeholder: torch.Tensor,
    ) -> None:
        """
        Internal use only
        """
        self._attn = attn
        self._lse = lse
        self._placeholder = placeholder

    def is_bmghk(self) -> bool:
        return self._attn.ndim == 5

    def apply(self, fn: Callable[[torch.Tensor], torch.Tensor]) -> "Partial":
        """
        Applies fn to the attention output, as if we were a tensor.
        fn must expect tensors of shape BMGHK or BMHK, but cannot actually
        manipulate the K dimension (because the LSE doesn't have one).

        For example, to slice on the sequence dimension, you might apply
        `lambda x: x[:, start:end]`.
        """
        attn = fn(self._attn)
        if self.is_bmghk():
            rearranged = self._lse.permute(0, 3, 1, 2).unsqueeze(-1)
            lse = fn(rearranged).squeeze(-1).permute(0, 2, 3, 1)
        else:
            rearranged = self._lse.permute(0, 2, 1).unsqueeze(-1)
            lse = fn(rearranged).squeeze(-1).permute(0, 2, 1)
        placeholder = fn(self._placeholder)
        return self.__class__(attn, lse, placeholder)

    def _tuple(self) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        return self._attn, self._lse, self._placeholder


def memory_efficient_attention_partial_autograd(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_bias: Optional[Union[torch.Tensor, fmha.AttentionBias]] = None,
    p: float = 0.0,
    scale: Optional[float] = None,
    *,
    op: Optional[Union[fmha.AttentionOp, Type[fmha.AttentionFwOpBase]]] = None,
    output_dtype: Optional[torch.dtype] = None,
) -> Partial:
    """
    Wrapper around `memory_efficient_attention_partial` which works with autograd.
    Arguments are the same as for `memory_efficient_attention_partial`.
    """
    return Partial(
        *_PartialFunc.apply(query, key, value, attn_bias, p, scale, op, output_dtype)
    )


def merge_attentions_autograd(
    *partials: Partial,
) -> torch.Tensor:
    """
    Wrapper around merge_attentions which works with autograd.
    """
    args = [i for part in partials for i in part._tuple()]
    if len(args) == 0:
        raise ValueError("No partials to merge")
    return _MergeFunc.apply(*args)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import os
from typing import List, Tuple

import torch

from .. import _is_triton_available


# Copied over from the sequence parallel fused ops.
def _should_use_triton(device: torch.device, dtype: torch.dtype) -> bool:
    if not int(os.getenv("XFORMERS_TILED_MATMUL_ENABLE_TRITON", "1")):
        return False
    if not _is_triton_available():
        return False
    device_capability = torch.cuda.get_device_capability(device)
    # Triton seems to be having issues on P100 and V100 GPUs, such as
    # https://github.com/openai/triton/issues/1609
    # https://github.com/openai/triton/issues/1610
    # https://github.com/openai/triton/issues/1257#issuecomment-1532616965
    # and, in recent Triton versions (Jan 2024), returning wrong values.
    if device_capability < (8, 0):
        return False
    return True


def check_inputs(
    a: List[List[torch.Tensor]],
    b: List[List[torch.Tensor]],
) -> Tuple[List[int], List[int], List[int]]:
    assert len(a) >= 1 and len(a[0]) >= 1 and all(len(row) == len(a[0]) for row in a), (
        "the first operand must be a non-empty two-dimensional regular list of lists "
        "of tenors"
    )
    assert len(b) >= 1 and len(b[0]) >= 1 and all(len(row) == len(b[0]) for row in b), (
        "the second operand must be a non-empty two-dimensional regular list of lists "
        "of tenors"
    )

    m_tiles = len(a)
    k_tiles = len(a[0])
    assert len(b) == k_tiles, (
        "the first operand's inner dimension must match the second operand's outer "
        f"dimension, got {k_tiles} and {len(b)}"
    )
    n_tiles = len(b[0])

    ms = [a[tile_m][0].shape[0] for tile_m in range(m_tiles)]
    ns = [b[0][tile_n].shape[1] for tile_n in range(n_tiles)]
    aks = [a[0][tile_k].shape[1] for tile_k in range(k_tiles)]
    bks = [b[tile_k][0].shape[0] for tile_k in range(k_tiles)]

    for tile_m in range(m_tiles):
        for tile_k in range(k_tiles):
            assert a[tile_m][tile_k].shape[0] == ms[tile_m], (
                f"the tensors on row {tile_m} of the first operand must all have the "
                f"same size along the m dimension, got {ms[tile_m]} at position 0 and "
                f"{a[tile_m][tile_k].shape[0]} at position {tile_k}"
            )
            assert a[tile_m][tile_k].shape[1] == aks[tile_k], (
                f"the tensors on column {tile_k} of the first operand must all have "
                f"the same size along the k dimension, got {aks[tile_k]} at position 0 "
                f"and {a[tile_m][tile_k].shape[1]} at position {tile_m}"
            )

    for tile_n in range(n_tiles):
        for tile_k in range(k_tiles):
            assert b[tile_k][tile_n].shape[0] == bks[tile_k], (
                f"the tensors on row {tile_k} of the second operand must all have the "
                f"same size along the k dimension, got {bks[tile_k]} at position 0 and "
                f"{b[tile_k][tile_n].shape[0]} at position {tile_n}"
            )
            assert b[tile_k][tile_n].shape[1] == ns[tile_n], (
                f"the tensors on column {tile_n} of the second operand must all have "
                f"the same size along the n dimension, got {ns[tile_n]} at position 0 "
                f"and {b[tile_k][tile_n].shape[1]} at position {tile_k}"
            )

    for tile_k in range(k_tiles):
        assert aks[tile_k] == bks[tile_k], (
            f"the tensors on column {tile_k} of the first operand and those on row "
            f"{tile_k} of the second operand must have the same size along the k "
            f"dimension, got {aks[tile_k]} and {bks[tile_k]}"
        )
    ks = aks

    return ms, ns, ks


def check_output(out: List[List[torch.Tensor]], ms: List[int], ns: List[int]) -> None:
    m_tiles, n_tiles = len(ms), len(ns)
    assert (
        len(out) >= 1
        and len(out[0]) >= 1
        and all(len(row) == len(out[0]) for row in out)
    ), "out must be a non-empty two-dimensional regular list of lists of tenors"
    assert len(out) == m_tiles
    assert len(out[0]) == n_tiles
    cms = [out[tile_m][0].shape[0] for tile_m in range(m_tiles)]
    cns = [out[0][tile_n].shape[1] for tile_n in range(n_tiles)]
    for tile_m in range(m_tiles):
        for tile_n in range(n_tiles):
            assert out[tile_m][tile_n].shape[0] == cms[tile_m], (
                f"the tensors on row {tile_m} of out must all have the same size "
                f"along the m dimension, got {cms[tile_m]} at position 0 and "
                f"{out[tile_m][tile_n].shape[0]} at position {tile_n}"
            )
            assert out[tile_m][tile_n].shape[1] == cns[tile_n], (
                f"the tensors on column {tile_n} of out must all have the same "
                f"size along the k dimension, got {cns[tile_n]} at position 0 and "
                f"{out[tile_m][tile_n].shape[1]} at position {tile_m}"
            )
    for tile_m in range(m_tiles):
        assert cms[tile_m] == ms[tile_m], (
            f"the tensors on row {tile_m} of out and those on row {tile_m} of the "
            f"first operand must have the same size along the m dimension, got "
            f"{cms[tile_m]} and {ms[tile_m]}"
        )
    for tile_n in range(n_tiles):
        assert cns[tile_n] == ns[tile_n], (
            f"the tensors on column {tile_n} of out and those on column {tile_n} "
            f"of the second operand must have the same size along the n dimension, "
            f"got {cns[tile_n]} and {ns[tile_n]}"
        )


# Using out= args in PyTorch is complicated, especially with custom_ops and
# torch.compile (we need to declare our side-effects with mutates_args, which
# then requires a functionalization step, ...). Thus this out= variant of the
# operator is exposed as a PLAIN PYTHON function, and is not compilable nor
# differentiable. It needs to be invoked from within a custom_op elsewhere.
def tiled_matmul_out(
    a: List[List[torch.Tensor]],
    b: List[List[torch.Tensor]],
    out: List[List[torch.Tensor]],
) -> None:
    ms, ns, ks = check_inputs(a, b)
    check_output(out, ms, ns)

    # TODO We can try merging tiles that come from contiguous memory, using
    # stack_or_none, to further improve performance.

    # Because the Triton kernel is hardcoded for maximum three tiles.
    # Because, in turn, we aimed this at the fusion of wq/wk/wv.
    if (
        len(ms) <= 3
        and len(ks) <= 3
        and len(ns) <= 3
        and _should_use_triton(a[0][0].device, a[0][0].dtype)
    ):
        from ._triton.tiled_matmul_kernels import _launch_triton_matmul

        _launch_triton_matmul(a, b, out, ms, ns, ks)
    else:
        for tile_m in range(len(ms)):
            for tile_n in range(len(ns)):
                torch.mm(a[tile_m][0], b[0][tile_n], out=out[tile_m][tile_n])
                for tile_k in range(1, len(ks)):
                    out[tile_m][tile_n].addmm_(a[tile_m][tile_k], b[tile_k][tile_n])


def _flatten(x: List[List[torch.Tensor]], rows: int, cols: int) -> List[torch.Tensor]:
    assert len(x) == rows
    assert all(len(row) == cols for row in x)
    flat_x = [elem for row in x for elem in row]
    assert len(flat_x) == rows * cols
    return flat_x


def _unflatten(
    flat_x: List[torch.Tensor], rows: int, cols: int
) -> List[List[torch.Tensor]]:
    assert len(flat_x) == cols * rows
    x = [
        flat_x[row_offset : row_offset + cols]
        for row_offset in range(0, rows * cols, cols)
    ]
    assert len(x) == rows
    assert all(len(row) == cols for row in x)
    return x


def _flattened_transpose(
    flat_x: List[torch.Tensor], rows: int, cols: int
) -> List[torch.Tensor]:
    x = _unflatten(flat_x, rows, cols)
    transposed_x = [[elem.t() for elem in col] for col in zip(*x)]
    flat_transposed_x = _flatten(transposed_x, cols, rows)
    return flat_transposed_x


# PyTorch (custom_op and torch.compile, but also the dispatcher in general)
# have a hard time with Tensor[][] args. Thus we flatten them into Tensor[] to
# pass them into and out of this operator.
# See: https://github.com/pytorch/pytorch/issues/113022
@torch.library.custom_op(
    "xformers_python::tiled_matmul_fwd",
    mutates_args=(),
    device_types="cuda",
)
def tiled_matmul_fwd(
    flat_a: List[torch.Tensor],
    flat_b: List[torch.Tensor],
    ms: List[int],
    ns: List[int],
    ks: List[int],
) -> List[torch.Tensor]:
    a = _unflatten(flat_a, len(ms), len(ks))
    b = _unflatten(flat_b, len(ks), len(ns))

    c = [[a[0][0].new_empty((m, n)) for n in ns] for m in ms]
    tiled_matmul_out(a, b, out=c)

    return _flatten(c, len(ms), len(ns))


@torch.library.register_fake("xformers_python::tiled_matmul_fwd")
def tiled_matmul_fwd_fake(
    flat_a: List[torch.Tensor],
    flat_b: List[torch.Tensor],
    ms: List[int],
    ns: List[int],
    ks: List[int],
) -> List[torch.Tensor]:
    c = [[flat_a[0][0].new_empty((m, n)) for n in ns] for m in ms]
    return _flatten(c, len(ms), len(ks))


def tiled_matmul_setup_context(ctx, inputs, output):
    flat_a, flat_b, ctx.ms, ctx.ns, ctx.ks = inputs
    ctx.save_for_backward(*flat_a, *flat_b)


def tiled_matmul_bwd(ctx, flat_grad_c):
    assert len(ctx.saved_tensors) == len(ctx.ms) * len(ctx.ks) + len(ctx.ks) * len(
        ctx.ns
    )
    flat_a = ctx.saved_tensors[: len(ctx.ms) * len(ctx.ks)]
    flat_b = ctx.saved_tensors[-len(ctx.ks) * len(ctx.ns) :]

    flat_transposed_a = _flattened_transpose(flat_a, len(ctx.ms), len(ctx.ks))
    flat_transposed_b = _flattened_transpose(flat_b, len(ctx.ks), len(ctx.ns))

    flat_grad_a = tiled_matmul_fwd(
        flat_grad_c, flat_transposed_b, ctx.ms, ctx.ks, ctx.ns
    )
    flat_grad_b = tiled_matmul_fwd(
        flat_transposed_a, flat_grad_c, ctx.ks, ctx.ns, ctx.ms
    )

    return flat_grad_a, flat_grad_b, None, None, None


torch.library.register_autograd(
    "xformers_python::tiled_matmul_fwd",
    tiled_matmul_bwd,
    setup_context=tiled_matmul_setup_context,
)


def tiled_matmul(
    a: List[List[torch.Tensor]],
    b: List[List[torch.Tensor]],
) -> List[List[torch.Tensor]]:
    """Multiply two matrices given as grids of tiles

    It performs the matmul between A and B, which are given as two-dimensional
    grids of tiles (i.e., blocks), represented as lists of lists of tensors.
    The output will itself be a matrix in such a form. Formally:

        out[m][n] = sum(a[m][k] @ b[k][n] for k in range(...))

    with the obvious constraints needed to make it work, in terms of number of
    tiles and sizes of each tile.

    The interest of this operator is to improve performance by avoding wave
    quantization effects when doing independent matrix multiplications in
    series. Sometimes, when these matmuls have one operand in common, this can
    also be addressed by concatenating the other operands into a single matrix,
    and issuing a single matmul. However this isn't always possible (e.g., might
    break the checkpoint format) and it's an anti-pattern, as it obscures the
    logic (e.g., changing the modelling code out of performance reasons). This
    tiled matmul performs the same computation as if the matrices were merged,
    without merging them, simply through a smarter memory addressing scheme.

    The tiled matmul is less generic than a grouped matmul, which can also help
    with wave quantization, and doesn't need the matmuls to have the same lhs
    or rhs operand. However, a grouped matmul will write the result of each
    matmul to a separate output matrix, whereas the tiled matmul allows to add
    them together into a single output. This is needed during the backward pass
    of a linear layer, and it's the reason we wrote this instead of using a
    grouped matmul.

    The tiled matmul is implemented using a custom Triton kernel, which puts
    constraints on the strides of the tiles. All rows of A must have the same
    K stride, all columns of A must have the same M stride, and so on.

    Currently the tiled matmul supports at most three tiles on each dimension,
    although fewer can also be given. This is because we needed it to fuse the
    query, key and value weights of an attention layer. This limit can be
    increased if needed.

    This operator is differentiable.

    """
    # Inputs are checked inside the op, but we check them as well to make sure
    # that they are "regular" and can be flattened.
    ms, ns, ks = check_inputs(a, b)
    flat_a = _flatten(a, len(ms), len(ks))
    flat_b = _flatten(b, len(ks), len(ns))
    flat_c = tiled_matmul_fwd(flat_a, flat_b, ms, ns, ks)
    c = _unflatten(flat_c, len(ms), len(ns))
    return c
